"""
Configuration functionality and parsing
"""
import dataclasses
from os.path import join, basename, abspath
from glob import glob

from o2tuner.io import make_dir, parse_yaml, exists_dir


@dataclasses.dataclass
class WorkDir:
    """
    Use this object to set the working directory globally
    """
    path = None


WORK_DIR = WorkDir()


def get_work_dir():
    return WORK_DIR.path


def set_work_dir(work_dir):
    WORK_DIR.path = abspath(work_dir)


def resolve_path(rel_cwd):
    return join(WORK_DIR.path, rel_cwd)


def load_config(config_path):
    return parse_yaml(config_path)


def get_done_dir():
    return join(WORK_DIR.path, "o2tuner_done")


def prepare_work_dir():
    make_dir(WORK_DIR.path)
    o2tuner_done_dir = get_done_dir()
    make_dir(o2tuner_done_dir)


def get_stages_done():
    done_dir = get_done_dir()
    done_files = glob(f"{done_dir}/DONE_*")
    done = []
    for done_file in done_files:
        with open(done_file, "r", encoding="utf-8") as this_done:
            cwd_rel = this_done.readline().strip()
            if not exists_dir(join(WORK_DIR.path, cwd_rel)):
                continue
        done.append(basename(done_file)[5:])
    return done


def set_stage_done(name, rel_cwd):
    done_dir = get_done_dir()
    done_file = join(done_dir, f"DONE_{name}")
    with open(done_file, "w", encoding="utf-8") as done:
        done.write(rel_cwd)



"""
Common helper functionality
"""
import sys

from o2tuner.log import Log

LOG = Log()


def annotate_trial(trial, key, value):
    """
    Annotate a trial object with user key-value pair
    """
    user_attributes = trial.user_attrs

    if key in user_attributes:
        LOG.error(f"This trial has annotation {key} already")
        sys.exit(1)

    trial.set_user_attr(key, value)



"""
Utility to run stages, bound to main entrypoint run action
"""
import sys
import argparse
from os.path import join, abspath, expanduser
from os import getcwd, chdir

from o2tuner.system import run_command, import_function_from_file
from o2tuner.optimise import optimise
from o2tuner.io import make_dir
from o2tuner.config import load_config, prepare_work_dir, get_stages_done, set_stage_done, set_work_dir, get_work_dir
from o2tuner.graph import create_graph_walker
from o2tuner.inspector import O2TunerInspector
from o2tuner.log import Log

OPTIMISATION_STAGE_FLAG = 0
EVALUATION_STAGE_FLAG = 1
USER_STAGE_FLAG = 2

LOG = Log()


def get_stage_work_dir(name, config):
    cwd_rel = config.get("cwd", name)
    return join(get_work_dir(), cwd_rel), cwd_rel


def get_optuna_config(name, config):
    optuna_config = {k: v for k, v in config.items() if k not in ("entrypoint", "objective", "script", "config")}
    if "study" not in optuna_config:
        optuna_config["study"] = {"name": name}
        return optuna_config
    if "name" not in optuna_config["study"]:
        optuna_config["study"]["name"] = name
    return optuna_config


def run_cmd_or_python(cwd, name, config):
    if "cmd" not in config and "python" not in config:
        LOG.error("Need either the cmd or python section in the config")
        return False
    if "python" in config:
        path = config["python"].get("file", None)
        entrypoint = config["python"].get("entrypoint", None)
        if not path or not entrypoint:
            LOG.error("Need path to python file and entrypoint function")
            return False
        user_config = config.get("config", {})
        # import function to be execuuted
        func = import_function_from_file(path, entrypoint)
        # cache current directory
        this_dir = getcwd()
        # change to this cwd and afterwards back
        chdir(cwd)
        if user_config:
            ret = func(user_config)
        else:
            ret = func()
        chdir(this_dir)
        return ret
    # We need to extract the return code here in a reasonable manner
    # It is not well defined to just get it via psutil.Process.wait() since it can happen
    # that the process is already done and return code can not even be found in the
    # system's process table anymore
    run_command(config["cmd"], cwd=cwd, log_file=config.get("log_file", f"{name}.log"), wait=True)
    return True


def run_optimisation(cwd, name, config):
    """
    Wrapper to run the optimisation.
    """
    func_name = config.get("entrypoint", config.get("objective", None))
    if not func_name or "file" not in config:
        LOG.error("Need path to python file and name of function entrypoint")
        return False
    func = import_function_from_file(config["file"], func_name)
    if not func:
        return False
    optuna_config = get_optuna_config(name, config)

    user_config = config.get("config", {})
    # cache current directory
    this_dir = getcwd()
    # change to this cwd and afterwards back
    chdir(cwd)
    ret = optimise(func, optuna_config, work_dir="./", user_config=user_config)
    chdir(this_dir)
    return ret


def run_inspector(cwd, config, stages_optimisation):
    """
    Wrapper to run the optimisation.
    """
    func_name = config.get("entrypoint", None)
    if not func_name or "file" not in config:
        LOG.error("Need path to python file and name of function entrypoint")
        return False
    func = import_function_from_file(config["file"], func_name)
    if not func:
        return False
    # which optimisations to load
    if "optimisations" not in config:
        LOG.error("Need key \"optimisations\" to know which optimisations to load")
        return False

    user_config = config.get("config", {})

    inspectors = []

    for optimisation in config["optimisations"]:
        if optimisation not in stages_optimisation:
            LOG.warning(f"Optimisation stage {optimisation} not defined, cannot construct inspector for that. Skip...")
            continue
        opt_config = stages_optimisation[optimisation]
        opt_cwd, _ = get_stage_work_dir(optimisation, opt_config)
        optuna_config = get_optuna_config(optimisation, opt_config)
        insp = O2TunerInspector()
        if not insp.load(optuna_config, opt_cwd, user_config):
            continue
        inspectors.append(insp)

    if not inspectors:
        LOG.warning("No O2TunerInspectors loaded, nothing to do")
        return False

    # cache current directory
    this_dir = getcwd()
    # change to this cwd and afterwards back
    chdir(cwd)
    ret = func(inspectors, user_config)
    chdir(this_dir)
    return ret


def run_stages(config, which_stages):  # noqa: C901 pylint: disable=too-many-branches
    """
    Run the stages defined in the config specified by the user
    Run all if nothing is specified
    """
    stages_user = config.get("stages_user", {})
    stages_optimisation = config.get("stages_optimisation", {})
    stages_evaluation = config.get("stages_evaluation", {})

    # Flag normal user stages with False to indicate that this is not an optimisation step
    stages = [(name, value, USER_STAGE_FLAG) for name, value in stages_user.items()]
    for name, value in stages_optimisation.items():
        # Instead, flag these stages with True to indicate these are optimisation steps
        stages.append((name, value, OPTIMISATION_STAGE_FLAG))
    for name, value in stages_evaluation.items():
        # Instead, flag these stages with True to indicate these are optimisation steps
        stages.append((name, value, EVALUATION_STAGE_FLAG))

    if not stages:
        LOG.warning("No stages found, nothing to do")
        return 0

    # For quick mapping of name to ID
    stages_name_to_id = {stage[0]: i for i, stage in enumerate(stages)}

    # nodes just numbers from 0 to N_nodes - 1
    # edges are 2-tuples of node IDs, (origin, target)
    edges = []
    for target, stage in enumerate(stages):
        if "deps" not in stage[1]:
            continue
        for dep in stage[1]["deps"]:
            origin = stages_name_to_id[dep]
            edges.append((origin, target))

    # get the object that walks us through the dependencies
    walker = create_graph_walker(len(stages), edges)
    stages_done = get_stages_done()
    if not which_stages:
        which_stages = []
    for name in which_stages:
        # These will always have precedence over "done" status
        walker.set_to_do(stages_name_to_id[name])
    for name in stages_done:
        walker.set_done(stages_name_to_id[name])

    stages_to_do = walker.compute_topology()
    LOG.info("STAGES to do (in this order):")
    for std in stages_to_do:
        LOG.info(f"  -> {stages[std][0]}")
    LOG.info("LET'S GO\n")

    # Now loop stage-by-stage
    for ind in stages_to_do:
        name, value, stage_flag = stages[ind]
        LOG.info(f"--> STAGE {name} <--")
        cwd, cwd_rel = get_stage_work_dir(name, value)
        make_dir(cwd)

        if stage_flag == OPTIMISATION_STAGE_FLAG and not run_optimisation(cwd, name, value):
            LOG.error(f"There was a problem in optimisation stage: {name}")
            return 1
        if stage_flag == USER_STAGE_FLAG and not run_cmd_or_python(cwd, name, value):
            LOG.error(f"There was a problem in custom stage: {name}")
            return 1
        if stage_flag == EVALUATION_STAGE_FLAG and not run_inspector(cwd, value, stages_optimisation):
            LOG.error(f"There was a problem in evaluation stage: {name}")
            return 1
        walker.set_done(ind)
        set_stage_done(name, cwd_rel)

    return 0


def run(args):
    """
    arparse will execute this function
    """
    set_work_dir(abspath(expanduser(args.work_dir)))
    # This will already fail if the config is found not to be sane
    config = load_config(args.config)
    # prepare the working directory
    prepare_work_dir()
    # run all stages
    return run_stages(config, args.stages)


def main():

    parser = argparse.ArgumentParser(description="o2tuner optimisation entry point")
    parser.set_defaults(func=run)
    parser.add_argument("-c", "--config", help="your configuration", required=True)
    parser.add_argument("-d", "--work-dir", dest="work_dir", help="The working directory to run in", required=True)
    parser.add_argument("--stages", nargs="*", help="run until this stage")
    args = parser.parse_args()
    return args.func(args)


if __name__ == "__main__":
    sys.exit(main())



"""
Mainstream and custom backends should be placed here
"""
import sys
from math import sqrt, ceil
from collections import OrderedDict

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mplc
import matplotlib.cm as mplcm
import matplotlib.colorbar as mplcb
import seaborn as sns
import pandas as pd

from optuna.importance import get_param_importances
from optuna.study._study_direction import StudyDirection

from o2tuner.io import parse_yaml
from o2tuner.backends import load_or_create_study
from o2tuner.sampler import construct_sampler
from o2tuner.log import Log

LOG = Log()


class O2TunerInspector:
    """
    Inspect optuna study
    """

    def __init__(self):
        """
        Constructor
        """
        self._study = None
        self._importances = None
        self._opt_user_config = None

    def load(self, opt_config=None, opt_work_dir=None, opt_user_config=None):
        """
        Loading wrapper
        """
        if not opt_config and not opt_work_dir:
            LOG.warning("Nothing to load, no configuration given")
            return False
        if isinstance(opt_config, str):
            opt_config = parse_yaml(opt_config)
        if not opt_config:
            opt_config = {}
        sampler = construct_sampler(opt_config.get("sampler", None))
        storage = opt_config.get("study", {})
        _, self._study = load_or_create_study(storage.get("name", None), storage.get("storage", None), sampler, opt_work_dir)
        self._opt_user_config = opt_user_config
        importances = get_param_importances(self._study, evaluator=None, params=None, target=None)
        self._importances = OrderedDict(reversed(list(importances.items())))
        return True

    def get_annotation_per_trial(self, key, accept_missing_annotation=True):
        """
        Assemble history of requested annotation
        """
        if accept_missing_annotation:
            return [t.user_attrs[key] if key in t.user_attrs else None for t in self._study.trials]
        ret_list = []
        for trial in self._study.trials:
            user_attrs = trial.user_attrs
            if key not in user_attrs:
                LOG.error(f"Key {key} not in trial number {trial.number}.")
                sys.exit(1)
            ret_list.append(user_attrs[key])
        return ret_list

    def get_losses(self):
        """
        Simply return list of losses
        """
        return [t.value for t in self._study.trials]

    def get_most_important(self, n_most_important=20):
        importance_values = list(self._importances.values())
        param_names = list(self._importances.keys())
        n_most_important = min(n_most_important, len(self._importances))
        importance_values = importance_values[-n_most_important:]
        param_names = param_names[-n_most_important:]
        return param_names, importance_values

    def plot_importance(self, *, n_most_important=50, map_params=None):
        """
        Plot the importance of parameters
        Most of it based on https://optuna.readthedocs.io/en/stable/_modules/optuna/visualization/_param_importances.html#plot_param_importances

        However, add some functionality we would like to have here
        """
        LOG.info("Plotting importance")
        param_names, importance_values = self.get_most_important(n_most_important)

        if map_params:
            param_names = [map_params[pn] if pn in map_params else pn for pn in param_names]

        figure, ax = plt.subplots(figsize=(30, 10))

        y_pos = [i for i, _ in enumerate(importance_values)]
        ax.barh(y_pos, importance_values)
        ax.set_yticks(y_pos, labels=param_names)
        ax.set_xlabel("parameter importance")

        return figure, ax

    def plot_parallel_coordinates(self, *, n_most_important=20, map_params=None):
        """
        Plot parallel coordinates. Each horizontal line represents a trial, each vertical line a parameter
        """
        LOG.info("Plotting parallel coordinates")
        params, _ = self.get_most_important(n_most_important)

        curves = [[] for _ in self._study.trials]
        losses = self.get_losses()
        skip_trials = []

        for i, trial in enumerate(self._study.trials):
            for param_key in params:
                if param_key not in trial.params:
                    skip_trials.append(i)
                    continue
                curves[i].append(trial.params[param_key])

        # re-map parameter names
        if map_params:
            params = [map_params[pn] if pn in map_params else pn for pn in params]

        # order trials by loss and prepare colorbar
        norm_colors = mplc.Normalize(vmin=min(losses), vmax=max(losses))
        # colorbar and sorting of losses reversed if needed
        cmap, reverse = (mplcm.get_cmap("Blues_r"), True) if self._study.direction == StudyDirection.MINIMIZE else (mplcm.get_cmap("Blues"), False)
        curves = [c for _, c in sorted(zip(losses, curves), reverse=reverse)]
        # make sure curves of best losses are plotted last and hence on top
        losses.sort(reverse=reverse)
        x_axis = list(range(len(params) + 1))
        figure, axes = plt.subplots(1, len(params) + 1, figsize=(30, 10))
        for i, (loss, curve) in enumerate(zip(losses, curves)):
            if i in skip_trials:
                continue
            for ax in axes[:-1]:
                ax.plot(x_axis[:-1], curve, c=cmap(norm_colors(loss)))
        for i, ax in enumerate(axes[:-1]):
            ax.set_xticks(x_axis[:-1], labels=params, rotation=45)
            ax.set_xlim(x_axis[i], x_axis[i+1])
            ylims = ax.get_ylim()
            ylims_diff = ylims[1] - ylims[1]
            y_low, y_up = (ylims[0] - 0.1 * ylims_diff, ylims[1] + 0.1 * ylims_diff)
            ax.set_ylim(y_low, y_up)
            ax.get_yaxis().set_ticks([y_low, y_up])
            # trick to hide horizontal axis
            ax.spines['bottom'].set_alpha(0)
            ax.spines['top'].set_alpha(0)

        mplcb.ColorbarBase(axes[-1], cmap="Blues_r", norm=norm_colors, label="loss", ticks=[min(losses), max(losses)], )
        figure.subplots_adjust(wspace=0)
        figure.suptitle("Parallel coordinates", fontsize=40)

        return figure, axes

    def plot_slices(self, *, n_most_important=21, map_params=None):
        LOG.info("Plotting slices")
        params, _ = self.get_most_important(n_most_important)

        n_rows = ceil(sqrt(len(params)))
        n_cols = n_rows
        if len(params) > n_rows**2:
            n_rows += 1

        losses = self.get_losses()
        figure, axes = plt.subplots(n_rows, n_cols, figsize=(50, 50))
        axes = axes.flatten()

        norm_colors = mplc.Normalize(vmin=0, vmax=len(losses) - 1)
        cmap = mplcm.get_cmap("Blues")
        # re-map parameter names
        params_labels = params
        if map_params:
            params_labels = [map_params[pn] if pn in map_params else pn for pn in params]

        for param_key, label, ax in zip(params[::-1], params_labels[::-1], axes):
            plot_this = True
            values = []
            for trial in self._study.trials:
                if param_key not in trial.params:
                    plot_this = False
                    break
                values.append(trial.params[param_key])
            if not plot_this:
                continue
            ax.scatter(values, losses, cmap=cmap, norm=norm_colors, c=range(len(losses)))
            ax.set_xlabel(f"value of parameter {label}", fontsize=30)
            ax.set_ylabel("loss", fontsize=30)
            ax.tick_params(labelsize=30)

        for ax in axes[len(params):]:
            ax.set_visible(False)

        figure.suptitle("Loss as function of param values", fontsize=40)

        return figure, axes

    def plot_correlations(self, *, n_most_important=20, map_params=None):
        """
        Plot correlation among parameters
        """
        LOG.info("Plotting parameter correlations")
        params, _ = self.get_most_important(n_most_important)
        params_labels = params
        if map_params:
            params_labels = [map_params[pn] if pn in map_params else pn for pn in params]

        param_values = []
        params_plot = []
        for param_key, label in zip(params[::-1], params_labels[::-1]):
            plot_this = True
            values = []
            for trial in self._study.trials:
                if param_key not in trial.params:
                    plot_this = False
                    break
                values.append(trial.params[param_key])
            if not plot_this:
                continue
            param_values.append(values)
            params_plot.append(label)

        df = pd.DataFrame(list(zip(*param_values)), columns=params_plot)
        corr = df.corr()
        # Generate a mask for the upper triangle
        mask = np.triu(np.ones_like(corr, dtype=bool))

        # Set up the matplotlib figure
        figure, ax = plt.subplots(figsize=(20, 20))

        # Generate a custom diverging colormap
        cmap = sns.diverging_palette(230, 20, as_cmap=True)

        # Draw the heatmap with the mask and correct aspect ratio
        sns.heatmap(corr, mask=mask, cmap=cmap, vmax=1, vmin=-1, center=0, square=True, linewidths=.5, cbar_kws={"shrink": .5})
        ax.tick_params(axis="x", rotation=90)
        figure.suptitle("Parameter correlations over trials", fontsize=40)

        return figure, ax

    def plot_pairwise_scatter(self, *, n_most_important=20, map_params=None):
        """
        Plot correlation among parameters
        """
        LOG.info("Plotting pair-wise scatter")
        params, _ = self.get_most_important(n_most_important)
        params_labels = params
        if map_params:
            params_labels = [map_params[pn] if pn in map_params else pn for pn in params]

        param_values = []
        params_plot = []
        for param_key, label in zip(params[::-1], params_labels[::-1]):
            plot_this = True
            values = []
            for trial in self._study.trials:
                if param_key not in trial.params:
                    plot_this = False
                    break
                values.append(trial.params[param_key])
            if not plot_this:
                continue
            param_values.append(values)
            params_plot.append(label)

        df = pd.DataFrame(list(zip(*param_values)), columns=params_plot)

        # Draw the heatmap with the mask and correct aspect ratio
        ax = sns.pairplot(df, height=1.8, aspect=1.8, plot_kws=dict(edgecolor="k", linewidth=0.5), diag_kind="kde", diag_kws=dict(shade=True),
                          corner=True)
        figure = ax.figure
        figure.set_figheight(30)
        figure.set_figwidth(30)
        figure.suptitle("Pair-wise scatter plot", fontsize=40)

        return figure, ax



"""
Construct and manage samplers
"""
import sys
# NOTE: There are more sampler implemented in optuna, however, let's start with these
from optuna.samplers import BaseSampler, GridSampler, RandomSampler, TPESampler, NSGAIISampler
from o2tuner.log import Log

LOG = Log()


SAMPLERS = {"base": BaseSampler,
            "grid": GridSampler,
            "random": RandomSampler,
            "tpe": TPESampler,
            "genetic": NSGAIISampler}


def construct_sampler(sampler_config=None):
    if not sampler_config:
        return TPESampler()
    name = sampler_config.get("name").lower()
    if name not in SAMPLERS:
        LOG.error(f"Unknwon sampler {name}")
        sys.exit(1)
    # NOTE Only YAMLable arguments can be used for this of course. We need to find a way to pass more complex things.
    #      E.g. some samplers take lambda function/callables etc.
    args = sampler_config.get("args", {})
    return SAMPLERS[name](**args)



"""
Useful classes to define and walk through directed acyclic graphs
"""

import sys
from o2tuner.log import Log

LOG = Log()


class GraphDAG:  # pylint: disable=too-few-public-methods
    """
    Class storing the configuration of a directed acyclic graph
    """
    def __init__(self, n_nodes, edges):
        """
        Init

        Args:
            n_nodes: int
                simply the number of nodes. Internally, the nodes will be range(n_nodes)
            edges: iter of 2-tuples
                iterable of edges with tuples (origin, target)
        """
        # Number of nodes
        self.n_nodes = n_nodes
        # edges
        self.edges = edges

        # Map nodes pointing towards a nodes
        self.from_nodes = [[] for _ in range(n_nodes)]
        # Map nodes to where this node points to
        self.to_nodes = [[] for _ in range(n_nodes)]
        # graph as 2D
        self.graph = [[False] * n_nodes for _ in range(n_nodes)]
        # Store the degree of incoming edges for each node (persistent)
        self.in_degree = [0] * self.n_nodes
        # Store the degree of outgoing edges for each node (persistent)
        self.out_degree = [0] * self.n_nodes
        # the index in the topology

        # Fill graph and in-degree
        for origin, target in self.edges:
            if self.graph[origin][target]:
                # We saw this edge already
                continue
            # Mark this edge as seen
            self.graph[origin][target] = True

            if origin > n_nodes or target > n_nodes or origin < 0 or target < 0:
                LOG.error(f"Found edge ({origin}, {target}) but nodes must be >= 0 and < {n_nodes}")
                sys.exit(1)
            self.from_nodes[target].append(origin)
            self.to_nodes[origin].append(target)
            self.in_degree[target] += 1
            self.out_degree[origin] += 1

        # Store at least one possible topology
        self.topology = []
        self.make_topology()

    def make_topology(self):
        """
        basically Kahn's algorithm to find a topology and to check for loops
        (which would then obviously not be a DAG)
        Returns True if sane and False if there is a cyclic path or an unknown node
        """
        # source nodes, no incoming edge
        in_degree = self.in_degree.copy()
        queue = [i for i, v in enumerate(in_degree) if not v]
        if not queue:
            LOG.error("There is no source node in the topology")
            return False

        counter = 0
        while queue:
            current = queue.pop(0)
            if current >= self.n_nodes or current < 0:
                LOG.error(f"Found an edge which node {current} but nodes are only valid from 0 to {self.n_nodes - 1}.")
                return False
            self.topology.append(current)
            for target in self.to_nodes[current]:
                in_degree[target] -= 1
                if not in_degree[target]:
                    queue.append(target)
            counter += 1
        if counter != self.n_nodes:
            LOG.error("There is at least one cyclic dependency.")
            return False
        return True


class GraphDAGWalker:
    """
    Find our way(s) through a DAG
    """
    def __init__(self, graph):
        """
        Init
        """
        self.graph = graph
        self.n_nodes = self.graph.n_nodes
        # Transient in degree which is updated whenever an origin node is done
        self.in_degree_transient = self.graph.in_degree.copy()
        # Flag done nodes
        self.done = []
        # As well as as a mask
        self.done_mask = [False] * self.n_nodes
        # mask those that must be done
        self.must_be_done = []
        self.must_be_done_mask = [False] * self.n_nodes
        # A transient topology
        self.topology_transient = None

    def get_can_be_done(self):
        """
        Simply get all nodes that are currently possible
        """
        return [i for i, idg in enumerate(self.in_degree_transient) if not idg and not self.done_mask[i]]

    def can_be_done(self, node):
        """
        Ask whether this node can be done
        """
        return not self.in_degree_transient[node]

    def set_done(self, node):
        """
        Report back here when a node is done
        """
        self.done.append(node)
        self.done_mask[node] = True

    def set_to_do(self, node):
        """
        Specify nodes to be done
        """
        self.must_be_done.append(node)
        self.must_be_done_mask[node] = True

    def find_shortest_path_to(self, targets, visited, done_mask, topology):
        """
        Recursively find shortest path to a target node
        """

        queue = targets.copy()
        while queue:
            next_target = queue.pop()
            if visited[next_target]:
                continue
            visited[next_target] = True
            topology.append(next_target)
            for origin in self.graph.from_nodes[next_target]:
                if not done_mask[origin]:
                    queue.append(origin)
        topology.reverse()

    def compute_topology(self):
        """
        Compute an efficient topology
        """
        # Nodes that should be done superseed the fact that they are potentially done already
        done_mask = self.done_mask.copy()
        done = []
        self.in_degree_transient = self.graph.in_degree.copy()
        # Sort according to topology
        must_be_done = [node for node in self.graph.topology if self.must_be_done_mask[node]]

        for node in self.done:
            if self.must_be_done_mask[node]:
                done_mask[node] = False
                continue
            done.append(node)

        for node in done:
            for target in self.graph.to_nodes[node]:
                self.in_degree_transient[target] -= 1

        self.topology_transient = []
        if not self.must_be_done:
            # If nothing specified to be done, just return the topology minus the tasks that were masked as done
            self.topology_transient = [n for n in self.graph.topology if not done_mask[n]]
            return self.topology_transient

        visited = [False] * self.n_nodes
        self.find_shortest_path_to(must_be_done, visited, done_mask, self.topology_transient)

        return self.topology_transient


def create_graph_walker(n_nodes, edges):
    """
    Directly create an object to walk through a GraphDAG
    """
    return GraphDAGWalker(GraphDAG(n_nodes, edges))



"""
Common I/O functionality
"""

import sys
from os.path import expanduser, isfile, isdir, exists, abspath, join
from os import makedirs, remove, listdir
from shutil import rmtree
import json
import yaml

from o2tuner.log import Log

LOG = Log()


def exists_file(path):
    """wrapper around python's os.path.isfile

    this has the possibility to add additional functionality if needed
    """
    return isfile(path)


def exists_dir(path):
    """wrapper around python's os.path.isdir

    this has the possibility to add additional functionality if needed
    """
    return isdir(path)


def make_dir(path):
    """create a directory
    """
    if exists(path):
        if exists_file(path):
            # if it exists and if that is actually a file instead of a directory, fail here...
            LOG.error(f"Attempted to create directory {path}. However, a file seems to exist there, quitting")
            sys.exit(1)
            # ...otherwise just warn.
        LOG.warning(f"The directory {path} already exists, not overwriting")
        return
    # make the whole path structure
    makedirs(path)


def remove_any_impl(to_be_removed, keep=None):

    if not exists(to_be_removed):
        return

    if isfile(to_be_removed):
        remove(to_be_removed)
        return

    if not keep:
        rmtree(to_be_removed)
        return

    # at this point we are sure that it is a directory and that certain things should be kept
    contains = [join(to_be_removed, ld) for ld in listdir(to_be_removed)]

    # figure out potential overlap between this path and things to be kept
    keep_this = []
    for con in contains:
        for k in keep:
            if not k.find(con):
                # means found at index 0
                keep_this.append(con)

    to_be_removed = [con for con in contains if con not in keep_this]

    for tbr in to_be_removed:
        remove_any_impl(tbr, keep_this)


def remove_dir(to_be_removed, keep=None):
    """
    remove a file or directory but keep certain things inside on request
    """
    to_be_removed = abspath(to_be_removed)
    if keep:
        keep = [join(to_be_removed, k) for k in keep]
    remove_any_impl(to_be_removed, keep)


########
# YAML #
########
class NoAliasDumperYAML(yaml.SafeDumper):
    """
    Avoid refs in dumped YAML
    """
    def ignore_aliases(self, data):
        return True


def parse_yaml(path):
    """
    Wrap YAML reading
    https://stackoverflow.com/questions/13518819/avoid-references-in-pyyaml
    """
    path = expanduser(path)
    try:
        with open(path, encoding="utf8") as in_file:
            return yaml.safe_load(in_file)
    except (OSError, IOError, yaml.YAMLError) as exc:
        LOG.error(f"ERROR: Cannot parse YAML from {path} due to\n{exc}")
        sys.exit(1)


def dump_yaml(to_yaml, path, *, no_refs=False):
    """
    Wrap YAML writing
    """
    path = expanduser(path)
    try:
        with open(path, 'w', encoding="utf8") as out_file:
            if no_refs:
                yaml.dump_all([to_yaml], out_file, Dumper=NoAliasDumperYAML)
            else:
                yaml.safe_dump(to_yaml, out_file)
    except (OSError, IOError, yaml.YAMLError) as eexc:
        LOG.error(f"ERROR: Cannot write YAML to {path} due to\n{eexc}")
        sys.exit(1)


########
# YAML #
########
def parse_json(filepath):
    """
    Wrap JSON reading, needed for interfacing with O2 cut configuration files
    """
    filepath = expanduser(filepath)
    if not exists_file(filepath):
        LOG.error(f"ERROR: JSON file {filepath} does not exist.")
        sys.exit(1)
    with open(filepath, "r", encoding="utf8") as config_file:
        return json.load(config_file)


def dump_json(to_json, path):
    """
    Wrap JSON writing, needed for interfacing with O2 cut configuration files
    """
    path = expanduser(path)
    with open(path, 'w', encoding="utf8") as config_file:
        json.dump(to_json, config_file, indent=2)



"""
Optimise hook, wrapping parallelisation
"""

from time import sleep
from math import floor
from multiprocessing import Process
import functools

from o2tuner.io import make_dir, parse_yaml
from o2tuner.backends import OptunaHandler
from o2tuner.sampler import construct_sampler
from o2tuner.log import Log

LOG = Log()


def optimise_run(objective, optuna_storage_config, sampler, n_trials, work_dir, user_config, run_serial):
    """
    Run one of those per job
    """
    handler = OptunaHandler(optuna_storage_config.get("name", None), optuna_storage_config.get("storage", None), work_dir, user_config, run_serial)
    handler.initialise(n_trials)
    handler.set_objective(objective)
    handler.set_sampler(sampler)
    handler.optimise()
    handler.finalise()
    return 0


def optimise(objective, optuna_config, *, work_dir="o2tuner_optimise", user_config=None):
    """
    This is the entry point function for all optimisation runs

    args and kwargs will be forwarded to the objective function
    """

    # read in the configurations, if string, assume to parse a YAML, otherwise it is assumed to be a dictionary
    if isinstance(optuna_config, str):
        optuna_config = parse_yaml(optuna_config)

    trials = optuna_config.get("trials", 100)
    jobs = optuna_config.get("jobs", 1)

    # investigate storage properties
    optuna_storage_config = optuna_config.get("study", {})

    run_serial = False
    if not optuna_storage_config.get("name", None) or not optuna_storage_config.get("storage", None):
        # we reduce the number of jobs to 1. Either missing the table name or the storage path will anyway always lead to a new study
        LOG.info("No storage provided, running only one job.")
        run_serial = True
        jobs = 1

    if trials < jobs:
        LOG.info(f"Attempt to do {trials} trials, hence reducing the number of jobs from {jobs} to {trials}")
        jobs = trials

    trials_list = floor(trials / jobs)
    trials_list = [trials_list] * jobs
    # add the left-over trials simply to the last job for now
    trials_list[-1] += trials - sum(trials_list)

    LOG.info(f"Number of jobs: {jobs}\nNumber of trials: {trials}")

    sampler = construct_sampler(optuna_config.get("sampler", None))

    make_dir(work_dir)

    procs = []
    for trial in trials_list:
        procs.append(Process(target=optimise_run, args=(objective, optuna_storage_config, sampler, trial, work_dir, user_config, run_serial)))
        procs[-1].start()
        sleep(5)

    while True:
        is_alive = any(p.is_alive() for p in procs)
        if not is_alive:
            break
        # We assume here that the optimisation might take some time, so we can sleep for a bit
        sleep(10)
    return True


def needs_cwd(func):
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        return func(*args, **kwargs)
    decorator.needs_cwd = True
    return decorator



"""
Common system tools
"""
import sys
from os.path import join, abspath
import importlib
import psutil

from o2tuner.log import Log

LOG = Log()

LOADED_MODULES = {}


def run_command(cmd, *, cwd="./", log_file=None, wait=True):
    """
    Prepare command and run
    """
    if log_file is None:
        log_file = "log.log"
    cmd = f"{cmd} >{log_file} 2>&1"
    LOG.info(f"Running command {cmd}")
    proc = psutil.Popen(["/bin/bash", "-c", cmd], cwd=cwd)
    if wait:
        proc.wait()
    return proc, join(cwd, log_file)


def load_file_as_module(path, module_name):
    """
    load path as module
    """
    lookup_key = abspath(path)
    if path in LOADED_MODULES:
        return LOADED_MODULES[lookup_key].__name__

    if module_name in sys.modules:
        LOG.error(f"Module name {module_name} already present, cannot load")
        sys.exit(1)
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    LOADED_MODULES[lookup_key] = module
    return module.__name__


def import_function_from_module(module_name, function_name):
    module = importlib.import_module(module_name)
    if not hasattr(module, function_name):
        LOG.error(f"Cannot find function {function_name} in module {module.__name__}")
        sys.exit(1)
    return getattr(module, function_name)


def import_function_from_file(path, function_name):
    module_name = load_file_as_module(path, path.replace("/", "_").replace(".", "_"))
    return import_function_from_module(module_name, function_name)



"""
Mainstream and custom backends should be placed here
"""
import sys
from time import time
from os import getcwd, chdir
from os.path import join
from inspect import signature
import pickle

import optuna

from o2tuner.io import make_dir, exists_file
from o2tuner.utils import annotate_trial
from o2tuner.log import Log

LOG = Log()


def make_trial_directory(trial):
    """
    Make a directory and attach to trial object as user attribute
    """
    user_attributes = trial.user_attrs
    if "cwd" in user_attributes:
        LOG.error(f"This trial has already a directory attached: {user_attributes['cwd']}")
        sys.exit(1)
    if "cwd" not in trial.study.user_attrs:
        LOG.error("This optimisation was not configured to run inside a directory. Please define a working directory.")
        sys.exit(1)
    top_dir = trial.study.user_attrs["cwd"]
    timestamp = str(int(time() * 1000000))
    cwd = join(top_dir, timestamp)
    make_dir(cwd)
    annotate_trial(trial, "cwd", cwd)
    return cwd


def adjust_path_sqlite(storage, workdir):
    """
    Make sure the path for SQLite is located either at an absolute path or relative to the specified workdir
    """
    if not storage:
        return None
    if storage.find("sqlite:///") != 0 or not workdir:
        return storage
    path = storage[10:]
    if path[0] == "/":
        # Absolute path
        return storage
    return "sqlite:///" + join(workdir, path)


def load_or_create_study(study_name=None, storage=None, sampler=None, workdir=None):
    """
    Helper to load or create a study
    Returns tuple of whether it can run via DB interface and optuna.study.study.Study
    """
    storage = adjust_path_sqlite(storage, workdir)
    if study_name and storage:
        # there is a database we can connect to for multiprocessing
        # Although optuna would come up with a unique name when study_name is None,
        # we force a name to be given by the user for those cases
        try:
            study = optuna.load_study(study_name=study_name, storage=storage, sampler=sampler)
            LOG.info(f"Loading existing study {study_name} from storage {storage}")
            return True, study
        except KeyError:
            study = optuna.create_study(study_name=study_name, storage=storage, sampler=sampler)
            LOG.info(f"Creating new study {study_name} at storage {storage}")
            return True, study
        except ImportError as exc:
            # Probably cannot import MySQL stuff
            LOG.info("Probably cannot import what is needed for database access. Will try to attempt a serial run.")
            LOG.info(exc)

    if study_name and workdir:
        # Try to see if there is a study saved here
        # Pickling is unsafe, we should try to find another way eventually
        file_name = join(workdir, f"{study_name}.pkl")
        if exists_file(file_name):
            with open(file_name, "rb") as save_file:
                LOG.info(f"Loading existing study {study_name} from file {file_name}")
                return False, pickle.load(save_file)

    return False, optuna.create_study(study_name=study_name, sampler=sampler)


def save_study(study, workdir):
    """
    Wrapper to pickle a study
    Pickling is unsafe, we should try to find another way eventually
    """
    if workdir:
        file_name = join(workdir, f"{study.study_name}.pkl")
        with open(file_name, "wb") as save_file:
            pickle.dump(study, save_file)


class OptunaHandler(object):
    """
    Handler based on Optuna backend
    """

    def __init__(self, db_study_name=None, db_storage=None, workdir=None, user_config=None, run_serial=False) -> None:
        """
        Constructor
        """
        # user objective function
        self._objective = None
        # Flag whether we need a dedicated cwd per trial
        self._needs_cwd_per_trial = False
        # chosen sampler (can be None, optuna will use TPE then)
        self._sampler = None
        # our study object
        self._study = None
        # number of trials to be done in given study
        self._n_trials = None
        # name of study (in database)
        self.db_study_name = db_study_name
        # database storage
        self.db_storage = db_storage
        # working directory (if any)
        self.workdir = workdir
        # optional user configuration that will be passed down to each call of the objective
        self.user_config = user_config
        # Flag to indicate if this is a serial run
        self.run_serial = run_serial

    def objective_wrapper(self, trial):
        cwd = None
        this_dir = getcwd()
        if self._needs_cwd_per_trial:
            cwd = make_trial_directory(trial)
        if cwd:
            chdir(cwd)
        ret = self._objective(trial)
        chdir(this_dir)
        return ret

    def initialise(self, n_trials=100):
        self._n_trials = n_trials
        has_db_access, self._study = load_or_create_study(self.db_study_name, self.db_storage, self._sampler, self.workdir)
        # Overwrite in case no DB access but a parallel execution was desired before
        self.run_serial = not has_db_access

        if self.workdir:
            self._study.set_user_attr("cwd", self.workdir)

    def optimise(self):
        if not self._n_trials or not self._objective:
            LOG.error("Not initialised: Number of trials and objective function need to be set")
            return
        self._study.optimize(self.objective_wrapper, n_trials=self._n_trials)

    def finalise(self):
        if self.run_serial and self.workdir:
            # Save the study if this is run serial and a workdir is given
            save_study(self._study, self.workdir)

    def set_objective(self, objective):
        sig = signature(objective)
        n_params = len(sig.parameters)
        if hasattr(objective, "needs_cwd"):
            self._needs_cwd_per_trial = True
        if n_params > 2 or not n_params:
            LOG.error("Invalid signature of objective funtion. Need either 1 argument (only trial obj) or 2 arguments (trial object + user_config)")
            sys.exit(1)
        if n_params == 1:
            self._objective = objective
        else:
            # Additionally pass the user config
            self._objective = lambda trial: objective(trial, self.user_config)

    def set_sampler(self, sampler):
        self._sampler = sampler



"""
o2tuner main module
"""

import sys
from pkg_resources import require
from o2tuner.argumentparser import O2TunerArgumentParser
from o2tuner.tuner import O2TunerError
from o2tuner.log import Log
from o2tuner.run import run


LOG = Log()


def entrypoint():
    arg_parser = O2TunerArgumentParser()
    # arg_parser.gen_config_help(O2Tuner.get_default_conf())
    args = arg_parser.parse_args()

    LOG.set_quiet(args.quiet)

    try:
        process_actions(args)
    except O2TunerError as exc:
        LOG.error(f"Cannot continue: {exc}")
        sys.exit(10)


def process_actions(args):
    if args.version:
        ver = str(require(__package__)[0].version)
        if ver == "LAST-TAG":
            ver = "development version"
        print(f"{__package__} {ver}")
        return

    if args.action in ["run"]:
        process_run(args)
    else:
        assert False, "invalid action"


def process_run(args):
    if args.action == "run":
        run(args)



"""
Module to process command line arguments
"""


import argparse


class O2TunerArgumentParser(argparse.ArgumentParser):
    """
    Lightweight subclass of ArgumentParser implementing some features required to display the
    o2tuner help: addition of normal arguments and matching between command-line
    options and configuration file options
    """

    def __init__(self):
        # self.args_normal = []
        super().__init__(formatter_class=argparse.RawTextHelpFormatter)
        super().add_argument("-v", "--version", dest="version", default=False, action="store_true",
                             help="Print current o2tuner version on stdout")
        super().add_argument("-w", "--workdir", dest="work_dir", help="Working directory to run in",
                             required=True)
        super().add_argument("-c", "--config", help="your configuration", required=True)
        super().add_argument("-q", "--quiet", dest="quiet", default=False,
                             action="store_true", help="Do not print any message")
        super().add_argument("-d", "--debug", dest="debug", default=None,
                             action="store_true", help="Increase verbosity level")
        super().add_argument("-s", "--stages", nargs="*", help="Run until specified stage")
        super().add_argument("action", default="run",
                             nargs="?", choices=["run", "init"], help="Actions to be performed")

    # def gen_config_help(self, default_conf):
    #     conf_file = os.path.join(os.getcwd(), ".o2tuner-config.yaml")
    #     epilog = f"It is possible to specify the most frequently used options in a YAML " \
    #              f"configuration file in your working directory\n" \
    #              f"Current expected path is: {conf_file}\n" \
    #              f"The following options (along with their default values) can be specified " \
    #              f"(please include `---` as first line):\n---\n"
    #     yaml_lines = {}
    #     longest = 0
    #     for opt in self.args_normal:
    #         if opt.config:
    #             assert opt.config in default_conf, f"option {opt.config} expected in default conf"
    #             optd = {opt.config: default_conf[opt.config]}
    #             yaml_lines[opt.option] = yaml.dump(
    #                 optd, default_flow_style=False).rstrip()
    #             longest = max(longest, len(yaml_lines[opt.option]))
    #     fmt = f"%%-{longest}s  # same as option %%s\n"
    #     for y_line in yaml_lines.items():
    #         epilog += fmt % (yaml_lines[y_line], y_line)

    #     self.epilog = epilog
